(in-package :xdump)

(defparameter *mode* (make-instance 'empire::xdump-mode :connection nil))

(defvar *current-parser* nil)

(defclass meta-table ()
  ((name :accessor meta-name)
   (type :accessor meta-type)
   (flags :accessor meta-flags)
   (len :accessor meta-len)
   (table :accessor meta-table)))

(defvar *meta-by-index* (make-array 30 :fill-pointer 0 :adjustable t))
(defvar *meta-index* (make-array 5))
(defvar *meta-meta* (make-array 5))

(defclass xdump-parser ()
  ((line-parser :accessor line-parser :initform nil)
   (entry-buffer :accessor entry-buffer :initform (make-array 1 :fill-pointer 0 :adjustable t))
   (entry-index :accessor entry-index :initform 0)
   (finalizer :accessor finalizer :initform nil)))

(defgeneric finish-table (xdump-parser number-of-records))
(defgeneric parse-entry (xdump-parser line))
(defgeneric parse-meta (xdump-parser name timestamp))
(defgeneric parse-table (xdump-parser name timestamp))
(defgeneric header-parser (xdump-parser line))
(defgeneric parse-line (xdump-parser line))

(defun meta-meta-finalizer (entry-buffer index)
  (declare (ignorable index))
  (loop
     for i = 0 then (+ i 1)
     for e across entry-buffer
     do (let* ((slot-name (string-upcase (car e)))
	       (slot (find-symbol slot-name :xdump)))
	  (setf (aref *meta-index* i) slot)))
  (loop for entry across entry-buffer
     for i = 0 then (+ i 1)
     do   (let ((meta-meta (make-instance 'meta-table)))
	    (loop for slot across *meta-index*
	       for field in entry
	       do (setf (slot-value meta-meta slot) field))
	    (setf (aref *meta-meta* i) meta-meta))))

(defun meta-finalizer (entry-buffer index)
  t)

(defmethod finish-table ((parser xdump-parser) number-of-records)
  (if (not (equal number-of-records (length (entry-buffer parser))))
      (error "Table row count mismatch"))
  (with-slots (finalizer entry-buffer entry-index) parser
    (funcall finalizer entry-buffer entry-index)))

(defmethod parse-entry ((parser xdump-parser) line)
  (if (char= #\/ (aref line 0))
      (finish-table parser (parse-integer (subseq line 1)))
      (let ((fields (mapcar #'read-from-string (cl-ppcre:split " " line))))
	(with-slots (entry-buffer entry-index) parser
	  (vector-push-extend fields entry-buffer)
	  (incf entry-index)))))

(defmethod parse-meta ((parser xdump-parser) name timestamp)
  (with-slots (line-parser finalizer) parser
    (setf line-parser #'parse-entry
	  finalizer (if (string= "meta" name)
			#'meta-meta-finalizer
			#'meta-finalizer))))

(defmethod parse-table ((parser xdump-parser) name timestamp)
  (setf (line-parser parser) #'parse-entry))

(defmethod header-parser ((parser xdump-parser) line)
  (multiple-value-bind (fullmatch groups)
      (cl-ppcre:scan-to-strings "^XDUMP meta (.*) ([0-9]+)$" line)
    (if fullmatch
	;; meta table
	(let ((name (aref groups 0))
	      (timestamp (parse-integer (aref groups 1))))
	  (parse-meta parser name timestamp))
	;; table
	(multiple-value-bind (fullmatch groups)
	    (cl-ppcre:scan-to-strings "^XDUMP (.*) ([0-9]+)$" line)
	  (declare (ignorable fullmatch))
	  (let ((name (aref groups 0))
		(timestamp (parse-integer (aref groups 1))))
	    (parse-table parser name timestamp))))))

(defmethod parse-line ((parser xdump-parser) line)
  (with-slots (line-parser) parser
    (funcall line-parser parser line)))

(defun make-parser ()
  (let ((parser (make-instance 'xdump-parser)))
    (setf (line-parser parser) #'header-parser)
    parser))

(defun parse (stream)
  (let ((parser (make-parser)))
    (setf *current-parser* parser)
    (loop for line = (read-line stream nil)
       while line do (parse-line parser line))))

(defun test-parse (filename)
  (with-open-file (s filename)
    (parse s)))

(defun xd-test (input)
  (loop for m in input
     do (empire::handle-data *mode* m)))

(defun t1 ()
  (test-parse "xdump-meta-meta.txt"))
